/*
 * Copyright (C) 2002 Tudor Bosman <tudorb-mutt@dwyn.net>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */ 

#include "mutt.h"
#include "imap.h"
#include "imap_private.h"
#include "imap_headercache.h"
#include "mx.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include <assert.h>

/* Delete all messages from headercache */
static int imap_headercache_purge(IMAP_HEADERCACHE *hc)
{
  int rc = -1;
  DIR *dir;
  struct dirent *ent;

  dir = opendir(hc->name);
  if (!dir)
  {
    mutt_error(_("IMAP headercache: can't purge directory %s: %s"), hc->name,
               strerror(errno));
    mutt_sleep(2);
    return -1;
  }

  while ((ent = readdir(dir)) != NULL)
  {
    if (!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, ".."))
      continue;

    sprintf(hc->tmpname, "%s/%s", hc->name, ent->d_name);
    if (unlink(hc->tmpname) == -1)
    {
      mutt_error(_("IMAP headercache: can't unlink file %s: %s"), hc->tmpname,
                 strerror(errno));
      mutt_sleep(2);
      goto bail;
    }
  }
  
  rc = 0;

bail:
  closedir(dir);

  return rc;
}

/* Open headercache */
IMAP_HEADERCACHE *imap_headercache_open(IMAP_DATA *idata)
{
  IMAP_HEADERCACHE *hc;
  char hcdir[_POSIX_PATH_MAX + 1];
  FILE *f;
  size_t len;
  char *p;

  if (!ImapHeadercache || ImapHeadercache[0] == '\0')
    return NULL;

  strfcpy(hcdir, ImapHeadercache, _POSIX_PATH_MAX);
  mutt_expand_path(hcdir, _POSIX_PATH_MAX);

  hc = safe_malloc(sizeof(IMAP_HEADERCACHE));

  len = strlen(hcdir) + strlen(idata->conn->account.host) +
        strlen(idata->mailbox) + 5;

  hc->name = safe_malloc(len);
  hc->tmpname = safe_malloc(len + NAME_MAX + 2);

  sprintf(hc->name, "%s/%s", hcdir, idata->conn->account.host);
  
  if (mkdir(hcdir, 0777) == -1 && errno != EEXIST)
  {
    mutt_error(_("Can't create IMAP headercache root directory %s: %s"), 
                 hcdir, strerror(errno));
    mutt_sleep(2);
    goto bail;
  }

  if (mkdir(hc->name, 0700) == -1 && errno != EEXIST)
  {
    mutt_error(_("Can't create IMAP headercache server directory %s: %s"),
                 hc->name, strerror(errno));
    mutt_sleep(2);
    goto bail;
  }

  p = idata->mailbox;
  while ((p = strchr(p, '/')) != NULL)
  {
    *p = '\0';
    sprintf(hc->name, "%s/%s/%s", hcdir, 
            idata->conn->account.host, idata->mailbox);

    if (mkdir(hc->name, 0700) == -1 && errno != EEXIST)
    {
      mutt_error(_("Can't create IMAP headercache mailbox directory %s: %s"),
                   hc->name, strerror(errno));
      mutt_sleep(2);
      goto bail;
    }

    *p = '/';
    p++;
  }

  sprintf(hc->name, "%s/%s/%s", hcdir, 
          idata->conn->account.host, idata->mailbox);

  if (mkdir(hc->name, 0700) == -1 && errno != EEXIST)
  {
    mutt_error(_("Can't create IMAP headercache mailbox directory %s: %s"),
                 hc->name, strerror(errno));
    mutt_sleep(2);
    goto bail;
  }
  
  sprintf(hc->tmpname, "%s/uidvalidity", hc->name);
  f = fopen(hc->tmpname, "r");

  if (f)
  {
    fscanf(f, "%u", &hc->uidvalidity);
    if (idata->uidvalidity != hc->uidvalidity)
    {
      fclose(f);
      f = NULL;
    }
  }

  if (!f)
  {
    if (imap_headercache_purge(hc) == -1)
      goto bail;

    sprintf(hc->tmpname, "%s/uidvalidity", hc->name);
    f = fopen(hc->tmpname, "w");
    if (!f)
    {
      mutt_error(_("Can't create IMAP headercache uidvalidity file %s: %s"),
                   hc->tmpname, strerror(errno));
      mutt_sleep(2);
      goto bail;
    }

    hc->uidvalidity = idata->uidvalidity;

    fprintf(f, "%u\n", hc->uidvalidity);
    fclose(f);
  }

  return hc;

bail:
  safe_free((void **)&hc->tmpname);
  safe_free((void **)&hc->name);
  safe_free((void **)&hc);

  return NULL;
}

/* Close headercache */
void imap_headercache_close(IMAP_HEADERCACHE *hc)
{
  safe_free((void **)&hc->tmpname);
  safe_free((void **)&hc->name);
  safe_free((void **)&hc);
}

static void imap_headercache_writehdr(FILE *f, IMAP_HEADER *h)
{
  /* Write the stuff in the header.  This must have a fixed length, as it is
   * overwritten in case of imap_headercache_update
   */
  fprintf(f, "%1x %1x %1x %1x %1x %1x %8x %16lx %16lx %8x\n",
          h->read, h->old, h->deleted, h->flagged, h->replied, h->changed,
          h->sid, h->received, h->content_length, HEADER_DATA(h)->uid);
}

/* Add message to headercache */
int imap_headercache_add(IMAP_HEADERCACHE *hc, IMAP_HEADER *h, FILE *from,
                         size_t hdrsz)
{
  FILE *f;
#define BUFSIZE 4096
  char buf[BUFSIZE];
  size_t sz;
  int rc = -1;

  sprintf(hc->tmpname, "%s/%u", hc->name, HEADER_DATA(h)->uid);

  f = fopen(hc->tmpname, "w");
  if (!f)
  {
    mutt_error(_("Can't create IMAP headercache message file %s: %s"),
                 hc->tmpname, strerror(errno));
    mutt_sleep(2);
    goto bail;
  }

  imap_headercache_writehdr(f, h);

  while ((sz = fread(buf, 1, (hdrsz < BUFSIZE ? hdrsz : BUFSIZE), from)) != 0)
  {
    hdrsz -= sz;
    fwrite(buf, 1, sz, f);
  }

  fclose(f);

  rc = 0;

bail:
  return rc;
}

/* Update flags in headercache message */
int imap_headercache_update(IMAP_HEADERCACHE *hc, IMAP_HEADER *h)
{
  FILE *f;
  int rc = -1;

  sprintf(hc->tmpname, "%s/%u", hc->name, HEADER_DATA(h)->uid);

  f = fopen(hc->tmpname, "r+");
  if (!f)
    goto bail;

  imap_headercache_writehdr(f, h);

  fclose(f);

  rc = 0;

bail:
  return rc;
}

/* Delete message from headercache */
int imap_headercache_delete(IMAP_HEADERCACHE *hc, IMAP_HEADER *h)
{
  int rc = -1;

  sprintf(hc->tmpname, "%s/%u", hc->name, HEADER_DATA(h)->uid);

  if (unlink(hc->tmpname) == -1)
  {
    mutt_error(_("Can't delete IMAP headercache message %s: %s"),
               hc->tmpname, strerror(errno));
    mutt_sleep(2);
    goto bail;
  }

  rc = 0;
  
bail:
  return rc;
}

/* Find message in headercache */
FILE *imap_headercache_find(IMAP_HEADERCACHE *hc, IMAP_HEADER *h)
{
  FILE *f = NULL;
  unsigned int flag_read, flag_old, flag_deleted, flag_flagged, flag_replied;
  unsigned int flag_changed;
  unsigned int uid;
  unsigned long received;
  unsigned long content_length;

  sprintf(hc->tmpname, "%s/%u", hc->name, HEADER_DATA(h)->uid);

  f = fopen(hc->tmpname, "r");
  if (!f)
    goto bail;

  fscanf(f, "%x %x %x %x %x %x %x %lx %lx %x\n",
         &flag_read, &flag_old, &flag_deleted, &flag_flagged, &flag_replied,
         &flag_changed, &h->sid, &received, &content_length, &uid);

  if (uid != HEADER_DATA(h)->uid)
  {
    fclose(f);
    f = NULL;
    goto bail;
  }

  h->received = received;
  h->read = flag_read;
  h->old = flag_old;
  h->deleted = flag_deleted;
  h->flagged = flag_flagged;
  h->replied = flag_replied;
  h->changed = flag_changed;
  h->content_length = (long)content_length;

bail:
  return f;
}

/* Close file returned by imap_headercache_find */
void imap_headercache_done(IMAP_HEADERCACHE *hc, FILE *f)
{
  fclose(f);
}

